# Part of the ROBOID project - http://hamster.school
# Copyright (C) 2016 Kwang-Hyun Park (akaii@kw.ac.kr)
# 
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# 
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General
# Public License along with this library; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330,
# Boston, MA  02111-1307  USA

from roboid.named_element import NamedElement


DeviceType = type('Enum', (), {'SENSOR': 0, 'EFFECTOR': 1, 'EVENT': 2, 'COMMAND': 3})
DataType = type('Enum', (), {'INTEGER': 4, 'FLOAT': 5, 'STRING': 6})


class Device(NamedElement):
    def __init__(self, id, name, device_type, data_type, data_size, min_value, max_value, initial_value):
        NamedElement.__init__(self, name)
        self.__id = id & 0xfff00fff
        self.__device_type = device_type
        self.__data_type = data_type
        self.__data_size = data_size
        self.__min_value = min_value
        self.__max_value = max_value
        self.__initial_value = initial_value
        if data_size > 0:
            self.__data = [initial_value] * data_size
        else:
            self.__data = []
        self.__event = False
        self.__fired = False
        self.__written = False
        self.__can_notify = False
        self.__device_data_changed_listeners = []

    def get_id(self):
        return self.__id

    def get_device_type(self):
        return self.__device_type

    def get_data_type(self):
        return self.__data_type

    def get_data_size(self):
        return self.__data_size

    def __is_written(self):
        return self.__written

    def __clear_written(self):
        self.__written = False

    def __check_data_type(self, value):
        if self.__data_type == DataType.INTEGER or self.__data_type == DataType.FLOAT:
            return isinstance(value, (int, float))
        elif self.__data_type == DataType.STRING:
            return isinstance(value, str)
        return False

    def __check_range(self, value):
        if isinstance(value, (int, float)):
            if value < self.__min_value:
                value = self.__min_value
            elif value > self.__max_value:
                value = self.__max_value
            if self.__data_type == DataType.INTEGER:
                return int(value)
            else:
                return float(value)
        return str(value)

    def e(self):
        return self.__event

    def read(self, arg=None):
        this_data = self.__data
        this_len = len(this_data)
        if isinstance(arg, (int, float)):
            index = int(arg)
            if index >= 0 and index < this_len:
                return this_data[index]
        elif isinstance(arg, list):
            length = min(this_len, len(arg))
            arg[:length] = this_data[:length]
            return length
        elif this_len > 0:
            return this_data[0]
        if self.__data_type == DataType.STRING:
            return ''
        else:
            return 0

    def write(self, arg1, arg2=None):
        if self.__device_type == DeviceType.SENSOR or self.__device_type == DeviceType.EVENT:
            return False
        this_data = self.__data
        if isinstance(arg1, (int, float)) and isinstance(arg2, (int, float, str)):
            index = int(arg1)
            if index < 0:
                return False
            if self.__check_data_type(arg2) == False:
                return False
            this_len = len(this_data)
            if index >= this_len:
                if self.__data_size < 0:
                    new_data = [self.__initial_value] * (index + 1)
                    new_data[:this_len] = this_data
                    this_data = self.__data = new_data
                else:
                    return False
            this_data[index] = self.__check_range(arg2)
            self.__fired = True
            self.__written = True
            self.__can_notify = True
            return True
        elif isinstance(arg1, (int, float, str)):
            if self.__check_data_type(arg1) == False:
                return False
            if len(this_data) <= 0:
                if self.__data_size < 0:
                    this_data = self.__data = [0]
                else:
                    return False
            this_data[0] = self.__check_range(arg1)
            self.__fired = True
            self.__written = True
            self.__can_notify = True
            return True
        elif isinstance(arg1, (list, tuple)):
            data_len = len(arg1)
            if data_len <= 0:
                return 0
            if self.__data_size < 0:
                if len(this_data) != data_len:
                    this_data = self.__data = [self.__initial_value] * data_len
            this_len = len(this_data)
            if this_len <= 0:
                return 0
            length = min(this_len, data_len)
            for i in range(length):
                if self.__check_data_type(arg1[i]):
                    this_data[i] = self.__check_range(arg1[i])
                    self.__fired = True
                    self.__written = True
                    self.__can_notify = True
            return length
        return False

    def __put(self, value):
        self.__data[0] = value
        self.__fired = True
        self.__can_notify = True

    def __put_at(self, index, value):
        self.__data[index] = value
        self.__fired = True
        self.__can_notify = True

    def __reset(self):
        self.__event = False
        self.__fired = False
        self.__written = False
        if self.__data_size > 0:
            self.__data = [self.__initial_value] * self.__data_size
        else:
            self.__data = []

    def add_device_data_changed_listener(self, listener):
        if listener:
            self.__device_data_changed_listeners.append(listener)

    def remove_device_data_changed_listener(self, listener):
        if listener:
            self.__device_data_changed_listeners.remove(listener)

    def clear_device_data_changed_listeners(self):
        self.__device_data_changed_listeners = []

    def __update_device_state(self):
        self.__event = self.__fired
        self.__fired = False

    def __notify_device_data_changed(self):
        if len(self.__device_data_changed_listeners) > 0:
            can_notify = True
            if self.__device_type == DeviceType.COMMAND or self.__device_type == DeviceType.EVENT:
                can_notify = self.__can_notify
            if can_notify:
                data = tuple(self.__data)
                for listener in self.__device_data_changed_listeners:
                    listener(self, data)
        self.__can_notify = False


class Roboid(NamedElement):
    def __init__(self, name, uid):
        NamedElement.__init__(self, name)
        self.__uid = uid
        self.__roboids = []
        self.__devices = []
        self.__sensory_devices = []
        self.__motoring_devices = []
        self.__disposed = False

    def __is_disposed(self):
        return self.__disposed

    def __dispose(self):
        self.__disposed = True

    def __reset(self):
        for device in self.__devices:
            device._Device__reset()
        for roboid in self.__roboids:
            roboid.__reset()

    def __add_device(self, id, name, device_type, data_type, data_size, min_value, max_value, initial_value):
        dev = Device(id, name, device_type, data_type, data_size, min_value, max_value, initial_value)
        self.__devices.append(dev)
        if device_type == DeviceType.SENSOR or device_type == DeviceType.EVENT:
            self.__sensory_devices.append(dev)
        else:
            self.__motoring_devices.append(dev)
        return dev

    def __add_roboid(self, roboid):
        if isinstance(roboid, Roboid):
            self.__roboids.append(roboid)

    def find_roboid_by_name(self, name):
        name = str(name).lower()
        dot = name.find('.')
        if dot < 0:
            for roboid in self.__roboids:
                if name == roboid.get_name().lower():
                    return roboid
        else:
            roboid_name = name[:dot]
            sub_name = name[dot+1:]
            for roboid in self.__roboids:
                if roboid_name == roboid.get_name().lower():
                    return roboid.find_roboid_by_name(sub_name)
        return None

    def find_device_by_name(self, name):
        name = str(name).lower()
        dot = name.find('.')
        if dot < 0:
            for device in self.__devices:
                if name == device.get_name().lower():
                    return device
        else:
            roboid_name = name[:dot]
            sub_name = name[dot+1:]
            for roboid in self.__roboids:
                if roboid_name == roboid.get_name().lower():
                    return roboid.find_device_by_name(sub_name)
        return None

    def add_device_data_changed_listener(self, listener):
        if listener:
            for device in self.__devices:
                device.add_device_data_changed_listener(listener)
            for roboid in self.__roboids:
                roboid.add_device_data_changed_listener(listener)

    def remove_device_data_changed_listener(self, listener):
        if listener:
            for device in self.__devices:
                device.remove_device_data_changed_listener(listener)
            for roboid in self.__roboids:
                roboid.remove_device_data_changed_listener(listener)

    def clear_device_data_changed_listeners(self):
        for device in self.__devices:
            device.clear_device_data_changed_listeners()
        for roboid in self.__roboids:
            roboid.clear_device_data_changed_listeners()

    def __update_sensory_device_state(self):
        for device in self.__sensory_devices:
            device._Device__update_device_state()
        for roboid in self.__roboids:
            roboid.__update_sensory_device_state()

    def __update_motoring_device_state(self):
        for device in self.__motoring_devices:
            device._Device__update_device_state()
        for roboid in self.__roboids:
            roboid.__update_motoring_device_state()

    def __notify_sensory_device_data_changed(self):
        for device in self.__sensory_devices:
            device._Device__notify_device_data_changed()
        for roboid in self.__roboids:
            roboid.__notify_sensory_device_data_changed()

    def __notify_motoring_device_data_changed(self):
        for device in self.__motoring_devices:
            device._Device__notify_device_data_changed()
        for roboid in self.__roboids:
            roboid.__notify_motoring_device_data_changed()

    def __clear_written(self):
        for device in self.__devices:
            device._Device__clear_written()
        for roboid in self.__roboids:
            roboid.__clear_written()


class Robot(NamedElement):
    def __init__(self, id, name, index, request_motoring_data):
        NamedElement.__init__(self, name)
        self.__id = id
        self.__index = index
        self.__request_motoring_data_callback = request_motoring_data
        self.__roboids = []

    def get_id(self):
        return self.__id

    def get_index(self):
        return self.__index

    def __add_roboid(self, roboid):
        if isinstance(roboid, Roboid):
            self.__roboids.append(roboid)

    def find_roboid_by_name(self, name):
        name = str(name).lower()
        dot = name.find('.')
        if dot < 0:
            for roboid in self.__roboids:
                if name == roboid.get_name().lower():
                    return roboid
        else:
            roboid_name = name[:dot]
            sub_name = name[dot+1:]
            for roboid in self.__roboids:
                if roboid_name == roboid.get_name().lower():
                    return roboid.find_roboid_by_name(sub_name)
        return None

    def find_device_by_name(self, name):
        name = str(name).lower()
        dot = name.find('.')
        if dot >= 0:
            roboid_name = name[:dot]
            sub_name = name[dot+1:]
            for roboid in self.__roboids:
                if roboid_name == roboid.get_name().lower():
                    return roboid.find_device_by_name(sub_name)
        return None

    def add_device_data_changed_listener(self, listener):
        if listener:
            for roboid in self.__roboids:
                roboid.add_device_data_changed_listener(listener)

    def remove_device_data_changed_listener(self, listener):
        if listener:
            for roboid in self.__roboids:
                roboid.remove_device_data_changed_listener(listener)

    def clear_device_data_changed_listeners(self):
        for roboid in self.__roboids:
            roboid.clear_device_data_changed_listeners()

    def __request_motoring_data(self):
        self.__request_motoring_data_callback()

    def __update_sensory_device_state(self):
        for roboid in self.__roboids:
            roboid._Roboid__update_sensory_device_state()

    def __update_motoring_device_state(self):
        for roboid in self.__roboids:
            roboid._Roboid__update_motoring_device_state()

    def __notify_sensory_device_data_changed(self):
        for roboid in self.__roboids:
            roboid._Roboid__notify_sensory_device_data_changed()

    def __notify_motoring_device_data_changed(self):
        for roboid in self.__roboids:
            roboid._Roboid__notify_motoring_device_data_changed()